#include "server.hpp"
#include "connexion.hpp"
#include "core.hpp"

#include "blc_channel.h"
#include "blc_program.h"

#include <vector>
#include <deque>
#include <map>
#include <thread>

using namespace std;

struct Client:Connexion{
    Client(blc_network *network);
    ~Client();
    
    thread *refresh_thread=nullptr;
    vector <blc_channel*> channels_to_refresh;
    
    
    void refresh_channels(period_us_t period_us);
    void send_channel_list();
    void send_channel_data(channel_id_t channel_id, period_us_t us_period);
    void send_channel_data(blc_channel *channel);

};

static deque<Client>clients; //We use deque to be sure Client will not be deleted and recreated during a reallocation (emplace).

Client::Client(blc_network *network):Connexion(network) {    
}

Client::~Client(){
    fprintf(stderr, "Disconnecting '%s:%s'\n", address_name, port_name);
}

void Client::send_channel_list(){
    blc_channel *tmp_channels=nullptr;
    int channels_nb;
    channel_id_t channel_id;
    char tmp_buffer[1024];
    size_t tmp_written;
    
    channels_nb=blc_channel_get_all_infos(&tmp_channels);
    channels.assign(tmp_channels, tmp_channels + channels_nb);
    FREE(tmp_channels);

    start_data(CHANNEL_DESCRIPTION);
    for(blc_channel &channel : channels) {
        fprintf(stderr, "channel %d\n", channel.id);
        channel_id=channel.id;
        WRITE_DATA(network->chars, write_pos, channel_id);
        tmp_written=channel.sprint_info(tmp_buffer, sizeof(tmp_buffer), true);
        bufferize_data(tmp_buffer, tmp_written);
    }
    fprintf(stderr, "%s", network->chars+9);
    end_data();
}



void Client::send_channel_data(blc_channel *channel){
    start_data(CHANNEL_DATA);
    WRITE_DATA(network->chars, write_pos, channel->id);
    bufferize_data(channel->chars, channel->size);
    end_data();
}

void Client::refresh_channels(period_us_t period_us){
    auto start = chrono::steady_clock::now();
    int diff_us;

    while(channels_to_refresh.size()){
        for(auto const &channel:channels_to_refresh){
            send_channel_data(channel);
        }
        auto end = chrono::steady_clock::now();
        diff_us=chrono::duration_cast<chrono::microseconds>(end-start).count();
        usleep(period_us-diff_us);
        start=end;
    }
}

void Client::send_channel_data(channel_id_t channel_id, period_us_t period_us){
    blc_channel *channel=nullptr;
    blc_channel *channel_to_refresh;
    
    for (blc_channel &tmp_channel : channels ) {
        fprintf(stderr, "channel %d\n", tmp_channel.id);
        if (tmp_channel.id == (int)channel_id) {
            channel = &tmp_channel;
            break;
        }
    }
    
    if (channel==nullptr) EXIT_ON_ERROR("Channel id '%d' not found.", channel_id);
    if (channel->fd==-1) channel->open(BLC_CHANNEL_READ);
    
    if (period_us==-1) { //We remove it from the list of channle to refresh and send it.
        try{
            channel_to_refresh=channels_to_refresh.at(channel_id);
        }
        catch (const out_of_range &except){
            channel_to_refresh=nullptr;
        }
        send_channel_data(channel);
    }
    else if (refresh_thread==nullptr) {
        refresh_thread=new thread(&Client::refresh_channels, this, period_us);
    }
}


static void on_receive_data(blc_server *server){
    Client *client=nullptr;
    uint8_t command;
    channel_id_t channel_id;
    period_us_t channel_period;
    
    for(Client &it : clients){
        if (it.remote_address_length==server->remote_address_length){
            if (memcmp(&it.remote_address, &server->remote_address, it.remote_address_length)==0){
                client=&it;
                break;
            }
        }
    }
    
    if (client==nullptr){
        clients.emplace_back(server);
        client=&clients.back();
        fprintf(stderr, "New connexion from '%s:%s'\n", client->address_name, client->port_name );
        client->first_read_header();
    }
    else client->read_header();
    READ_DATA(command, client->network->chars, client->read_pos);
    
    switch (command){
        case REQUEST_CHANNEL_LIST:
            fprintf(stderr, "List request from '%s:%s'\n", client->address_name, client->port_name) ;
            client->send_channel_list();
            break;
            
        case REQUEST_CHANNEL_DATA:
            READ_DATA(channel_id, client->network->chars, client->read_pos);
            READ_DATA(channel_period, client->network->chars, client->read_pos);
            fprintf(stderr, "Data request from '%s:%s' for channel '%d'\n", client->address_name, client->port_name, channel_id) ;
            client->send_channel_data(channel_id, channel_period);
            break;
            
        default: EXIT_ON_ERROR("Unknow command '%d'.", command);
    }
}

void start_server(char const *port){
    blc_server server(port, BLC_UDP4, on_receive_data, nullptr, -1);
    fprintf(stderr, "Server mode\n");
    
    BLC_COMMAND_LOOP(-1){
        
    }
    
}
